const child_process = require('child_process')
    , os = require('os')
    , path = require('path')
    , fse = require('fs-extra')
    , chokidar = require('chokidar')
    , EventEmitter = require('events')
    , escapeText = require('./_escape_text.js')
    , sleep = require('../utilities/sleep')
    , BdpAdapter = require('./adapter_class');

class PbsEventEmitter extends EventEmitter {}


const jsonifyQstat = function(obj){
    var results = {
        "jobId"     :   obj[0],
        "user"      :   obj[1],
        "queue"     :   obj[2],
        "name"      :   obj[3],
        "sessionID" :   obj[4],
        "NDS"       :   obj[5],
        "TSK"       :   obj[6],
        "req_memory":   obj[7],
        "req_time"  :   obj[8],
        "status"    :   obj[9],
        "time"      :   obj[10],
        "comment"   :   obj[11],
    };
    return results;
}
const checkJobExitCode = function(jobID) {
    return new Promise((resolve, reject) => {
        const qstatProcess = child_process.spawn('qstat', ['-x', '-f', jobID]);
        let jobInfo = '';
        qstatProcess.stdout.on('readable', () => {
            const data = qstatProcess.stdout.read();
            jobInfo += data;
        });
        qstatProcess.on('error', (err) => {
            reject(err);
        });
        qstatProcess.on('exit', (code, signal) => {
            if (code !== 0) {
                reject('Exit code: ' + code + '; Signal: ' + signal);
                return;
            }
            jobInfo = jobInfo.trim().replace(/\n\t/g,"").split('\n');
            const results = {};
            for (var i = 1; i < jobInfo.length; i++) {
                if (jobInfo[i].indexOf(' = ')!== -1){
                    // Split key and value to 0 and 1
                    var data = jobInfo[i].split(' = ');
                    results[data[0].trim()] = data[1].trim();
                }
            }
            resolve(results['Exit_status']);
        });
    });
}

const checkStatus = function(pbsUser, runningTasks) {
    const qstat = child_process.spawnSync('qstat', ['-x', '-w', '-u', pbsUser], {shell: true, encoding: 'utf8'});
    const lines = qstat.stdout.split('\n');
    const jobs = {};
    for (let i = 5; i < lines.length; i ++) {
        lines[i] = lines[i].replace(/^.*?\s/,function myFunction(jobname){return jobname + "  ";});
        lines[i] = lines[i].replace(/\s[A-Z]\s/,function myFunction(status){return "  " + status + "  ";});
        lines[i] = lines[i].trim().split(/[\s]{2,}/);
        jobs[lines[i][0]] = jsonifyQstat(lines[i]);
    }
    const latestStatus = {
        finished: 0,
        exit: 0,
        running: 0,
        queued: 0,
        others: []
    };
    const jobIDs = Object.keys(runningTasks);
    for (const jobID of jobIDs) {
        if (jobs[jobID]) {
            const status = jobs[jobID].status;
            switch(status) {
                case 'F':
                    runningTasks[jobID].emit('finish');
                    latestStatus.finished ++;
                break;
                case 'E':
                    runningTasks[jobID].emit('error');
                    latestStatus.exit ++;
                break;
                case 'R':
                    latestStatus.running ++;
                break;
                case 'Q':
                    latestStatus.queued ++;
                break;
                default:
                    runningTasks[jobID].emit('error');
                    latestStatus.others.push({status: status, jobId: jobID});
                break;
            }
        }else {
            runningTasks[jobID].emit('error');
        }
    }
    return latestStatus;
}



class BdpPbsAdapter extends BdpAdapter {
    constructor (adapterName, author, opt){
        super(adapterName, author, opt);
        this.userInfo = os.userInfo();
        this.homedir = os.homedir();
        this.sparkJobID = null;
        this.sparkMaster = null;
        this.sparkStatusPath = null;
    }
    async beforeStart() {
        const options = this.options;
        if (!options.sparkPath) {
            throw 'Please specify the spark folder path.';
        }
        const preProcess = child_process.spawnSync('qsub', [
            '-l'
            , 'select=' + options.nodes + ':ncpus=' + options.cpus + ':mem=' + options.mem
            , '-V'
            , '-k'
            , 'oe'
            , '-N'
            , 'bdp-spark-deploy'
            , '--'
            , '/bin/sh'
            , path.resolve(process.env.PACKAGE_FOLDER, 'node_modules', 'bdp-task-adapter', 'lib', 'adapters', 'pbs-spark', 'bdp-pbs-spark-deploy.sh', options.sparkPath)
        ]);
        this.sparkJobID = preProcess.stdout.replace(/\n/g, "");
        let sparkDeployed = false;
        this.sparkStatusPath = path.resolve(this.homedir, 'spark' + this.sparkJobID);
        while(!sparkDeployed) {
            sparkDeployed = await fse.pathExists(this.sparkStatusPath);
            await sleep(this.options.updateInterval);
        }
        const sparkStatus = await fse.readFile(this.sparkStatusPath, 'utf8');
        const sparkStatusArr = sparkStatus.split('\n');
        if (sparkStatusArr[0] !== "0") {
            throw 'Failed to deploy spark clusters.'
        }
        this.sparkMaster = sparkStatusArr[1];
        return true;
    }
    beforeExit() {
        const jobIDs = Object.keys(this.runningTasks);
        for (const jobID of jobIDs) {
            process.stderr.write('[' + new Date().toString() + '] Stopping jobID: ' + jobID +'\n');
            child_process.spawnSync('qdel', [jobID], {shell: true, encoding: 'utf8'});
            process.stderr.write('[' + new Date().toString() + '] Stopped\n');
            this.runningTasks[jobID].emit('error');
        }
        try {
            fse.removeSync(this.sparkStatusPath);
        }catch(e) {
            child_process.spawnSync('qdel', [this.sparkJobID], {shell: true, encoding: 'utf8'});
        }
    }
    async runProcess(taskObj, processID) {
        let stdoeWatcher;
        const options = this.options;
        const cwd = this.options.cwd || process.cwd();
        const runningTasks = this.runningTasks;
        const taskLogs = this.taskLogs;
        const args = taskObj.args;
        const startTime = new Date().valueOf();
        const readfileSize = {stdout: 0, stderr: 0};
        const stdoutFile = path.resolve(options.taskLogFolder, processID + '-stdout.txt');
        const stderrFile = path.resolve(options.taskLogFolder, processID + '-stderr.txt');
        const qsub = child_process.spawnSync('qsub', args, {cwd: cwd, shell: true, encoding: 'utf8', env: process.env});
        const jobID = qsub.stdout.replace(/\n/g,"");
        // const stdoutFS = fse.createWriteStream(stdoutFile);
        // const stderrFS = fse.createWriteStream(stderrFile);
        runningTasks[jobID] = new PbsEventEmitter();
        let msg = '[' + new Date(startTime).toString() + '] ' + taskObj.taskName + ' Starting.\n';
        process.stderr.write(msg);
        msg = '[command] qusb ' + args.join(' ') + '\n';
        process.stderr.write(msg);
        taskLogs[processID].pid = jobID;
        taskLogs[processID].start = startTime;
        taskLogs[processID].command = 'qsub ' + args.join(' ');
        const matched = jobID.match(/^(\d+)\..*$/);
        const runtimeStdOut = path.resolve(this.homedir, processID + '.o' + matched[1]);
        const runtimeStdErr = path.resolve(this.homedir, processID + '.e' + matched[1]);
        taskLogs[processID].stdout = runtimeStdOut;
        taskLogs[processID].stderr = runtimeStdErr;
        if (!options.batch) {
            console.log('Watching stadard outputs/errors.');
            stdoeWatcher = chokidar.watch(this.homedir + '/' + processID + '.*', {usePolling: true, ignoreInitial: true, followSymlinks: false, cwd: this.homedir, alwaysStat: true});
            stdoeWatcher.on('add', (thePath, stats) => {
                const stdoe = thePath === processID + '.o' + matched[1] ? 'stdout' : 'stderr';
                fse.createReadStream(path.resolve(this.homedir, thePath), {encoding: 'utf8', start: readfileSize[stdoe], end: stats.size}).pipe(process[stdoe]);
                readfileSize[stdoe] = stats.size;
            });
            stdoeWatcher.on('change', (thePath, stats) => {
                const stdoe = thePath === processID + '.o' + matched[1] ? 'stdout' : 'stderr';
                fse.createReadStream(path.resolve(this.homedir, thePath), {encoding: 'utf8', start: readfileSize[stdoe], end: stats.size}).pipe(process[stdoe]);
                readfileSize[stdoe] = stats.size;
            });
        }
        const getRemainingStdoe = function(stdoe, filePath, start) {
            return new Promise((resolve) => {
                const stdoeFS = fse.createReadStream(filePath, {encoding: 'utf8', start: start});
                stdoeFS.on('end', () => {
                    resolve();
                });
                stdoeFS.on('error', () => {
                    // ignore the error, currently.
                    resolve();
                });
                stdoeFS.pipe(process[stdoe]);
            });
        };
        const exitCallback = async () => {
            taskLogs[processID].end = new Date().valueOf();
            const stdoutExists = await fse.pathExists(runtimeStdOut);
            const stderrExists = await fse.pathExists(runtimeStdErr);
            if (!options.batch) {
                stdoeWatcher.close();
                if (stdoutExists) {
                    await getRemainingStdoe('stdout', runtimeStdOut, readfileSize.stdout);
                }
                if (stderrExists) {
                    await getRemainingStdoe('stderr', runtimeStdErr, readfileSize.stderr);
                }
            }
            delete runningTasks[jobID];
            if (stdoutExists) {
                await fse.move(runtimeStdOut, stdoutFile, {overwrite: true});
                taskLogs[processID].stdout = stdoutFile;
            }else {
                taskLogs[processID].stdout = null;
            }
            if (stderrExists) {
                await fse.move(runtimeStdErr, stderrFile, {overwrite: true});
                taskLogs[processID].stderr = stderrFile;
            }else {
                taskLogs[processID].stderr = null;
            }
            await this.updateStatus(true);
            const exitCode = await checkJobExitCode(jobID);
            taskLogs[processID].exitCode = exitCode;
            return exitCode;
        }
        await this.updateStatus(true);
        return new Promise((resolve, reject) => {
            if (qsub.stderr) {
                console.log(qsub.stderr);
                reject('Can not perform qsub\n');
            }
            runningTasks[jobID].on('finish', () => {
                exitCallback().then((exitCode) => {
                    exitCode == 0 ? resolve() : reject(exitCode);
                }).catch((err) => {
                    reject(err);
                });
            });
            runningTasks[jobID].on('error', () => {
                exitCallback().then((exitCode) => {
                    reject(exitCode);
                }).catch((err) => {
                    reject(err);
                });
            });
        });
    }
    addTasks(tasks) {
        const options = this.options;
        for (let i = 0; i < tasks.length; i ++) {
            const eachTask = tasks[i];
            const taskOption = Object.assign({}, options, eachTask.option);
            const processID = eachTask.name + '-' + new Date().getTime() + '-' + Math.floor(1000000*Math.random());
            let nodes = taskOption.nodes;
            if (isNaN(parseInt(nodes))) { nodes = taskOption.nodes; }else { nodes = parseInt(nodes);}
            let cpus = taskOption.cpus;
            if (isNaN(parseInt(cpus))) { cpus = taskOption.cpus; }else { cpus = parseInt(cpus); }
            let mem = eachTask.option && eachTask.option.mem ? eachTask.option.mem : options.mem;
            mem = mem.toLowerCase().match(/^(\d+[bkmg])/);
            mem = mem ? mem[1] + 'b' : options.mem;
            const preQsubArgs = ['-l'
                , 'select=' + nodes + ':ncpus=' + cpus + ':mem=' + mem
                , '-v'
                , 'PBS_O_WORKDIR=`pwd`'
                , '-V'
                , '-k'
                , 'oe'
                , '-N'
            ];
            const qsubArgs = [];
            qsubArgs.push.apply(qsubArgs, preQsubArgs);
            qsubArgs.push(processID);
            qsubArgs.push('--');
            if (eachTask.dockerImage) {
                qsubArgs.push('/bin/bash');
                qsubArgs.push('-c');
                const preDockerArgs = [];
                if (taskOption.imagePull) {
                    preDockerArgs.push.apply(preDockerArgs, [
                        taskOption.dockerPath
                        , 'pull'
                        , eachTask.dockerImage
                        , '&&'
                    ]);
                }
                preDockerArgs.push.apply(preDockerArgs, [taskOption.dockerPath, 'run', '--rm']);
                if (taskOption.dataFolder) {
                    preDockerArgs.push.apply(preDockerArgs, [
                        '-v', taskOption.dataFolder + ':/data', '-e', 'PROJECT_FOLDER=' + taskOption.dataFolder
                    ]);
                }
                if (taskOption.configFolder) {
                    preDockerArgs.push.apply(preDockerArgs, [
                        '-v', taskOption.configFolder + ':/config', '-e', 'PACKAGE_FOLDER=' + taskOption.configFolder
                    ]);
                }
                if (taskOption.workdir) {
                    preDockerArgs.push.apply(preDockerArgs, [
                      '--workdir'
                      , taskOption.workdir
                    ]);
                }
                if (os.platform() !== 'win32') {
                    if (!taskOption.setUser) {
                        preDockerArgs.push.apply(preDockerArgs, [
                            '-e'
                            , 'HOST_USER_ID=' + this.userInfo.uid
                            , '-e'
                            , 'HOST_USER_GROUP=' + this.userInfo.gid
                        ]);
                    }else {
                        preDockerArgs.push.apply(preDockerArgs, [
                            '--user'
                            , this.userInfo.uid + ':'+ this.userInfo.gid
                        ]);
                    }
                    if (Array.isArray(taskOption.extFolders)) {
                        for (let i = 0; i < taskOption.extFolders.length; i ++) {
                            preDockerArgs.push('-v');
                            preDockerArgs.push(taskOption.extFolders[i] + ':' + taskOption.extFolders[i]);
                        }
                    }
                }
                // qsubArgs.push(preDockerArgs.join(' ') + ' ' + task.dockerImage + ' ' + task.exec + ' ' + task.args.join(' '));
                const taskArgs = [];
                if (os.platform() !== 'win32') {
                    taskArgs.push.apply(taskArgs, eachTask.args.map((arg) => {
                        return '"\'' + escapeText(arg) + '\'"';
                    }));
                }else {
                    taskArgs.push.apply(taskArgs, eachTask.args);
                }
                qsubArgs.push( '\"' + preDockerArgs.join(' ') + ' ' + eachTask.dockerImage + ' ' + eachTask.exec + ' ' + taskArgs.join(' ') +'\"');
            }else {
                qsubArgs.push(eachTask.exec);
                qsubArgs.push.apply(qsubArgs, eachTask.args);
            }
            this.taskLogs[processID] = {
                taskName: eachTask.name
                , pid: -1
                , image: eachTask.dockerImage
                , exec: eachTask.exec
                , args: qsubArgs
                , command: null
                , start: null
                , end: null
                , stdout: ''
                , stderr: ''
                , cpus: cpus
                , mem: mem
                , nodes: 1
                , exitCode: null
            }
        }
    }
    printStatus(currentStatus) {
        process.stdout.write('[' + new Date().toString() + '] ' + 'status change detected.\n');
        let statusOutput = 
            'Pending: ' + (currentStatus.pending) + '\n'
            + 'Queued: ' + (currentStatus.queued) + '\n'
            + 'Running: ' + (currentStatus.running) + '\n'
            + 'Finishing: ' + (currentStatus.finished) + '\n'
            + 'Exiting: ' + (currentStatus.exit) + '\n'
            + 'Done: ' + (currentStatus.done) + '\n'
            + 'Total: ' + (currentStatus.total) + '\n';
        if (currentStatus.others.length > 0) {
            statusOutput += '-------------------------------\n';
            statusOutput += currentStatus.others.map((statusObj) => {
                return 'Job ' + statusObj.jobId + ': ' + statusObj.status + '\n';
            });
            statusOutput += '-------------------------------\n';
        }else {
            statusOutput += '-------------------------------\n';
        }
        process.stdout.write(statusOutput);
    }
    async monitorTasks(queue) {
        let previousStatus = {
            pending: 0,
            queued: 0,
            running: 0,
            finished: 0,
            exit: 0,
            done: 0,
            others: [],
            total: 0
        };
        while (Object.keys(this.runningTasks).length > 0 && queue.length > 0) {
            const jobIDs = Object.keys(this.runningTasks);
            const taskLogs = this.taskLogs;
            const currentStatus = checkStatus(this.userInfo.username, this.runningTasks);
            if (this.options.batch) {
                currentStatus.pending = Object.keys(taskLogs).filter((processID) => {
                    return taskLogs[processID].exitCode === null && jobIDs.indexOf(taskLogs[processID].pid) < 0;
                }).length;
                currentStatus.done = Object.keys(taskLogs).filter((processID) => {
                    return taskLogs[processID].exitCode !== null && jobIDs.indexOf(taskLogs[processID].pid) < 0;
                }).length;
                currentStatus.total = Object.keys(taskLogs).length;
                for (const state of Object.keys(currentStatus)) {
                    if ((state === 'others' && currentStatus[state].length !== previousStatus[state].length) || (state !== 'others' && currentStatus[state] !== previousStatus[state])) {
                        this.printStatus(currentStatus);
                        break;
                    }
                }
                previousStatus = currentStatus;
            }
            await sleep(this.options.updateInterval);
        }
    }
}

module.exports = BdpPbsAdapter;
