const child_process = require("child_process"),
  os = require("os"),
  path = require("path"),
  fse = require("fs-extra"),
  EventEmitter = require("events"),
  sleep = require("../../utilities/sleep"),
  memHumanize = require("../../utilities/humanize-memory"),
  spawnProcessAsync = require("../../utilities/spawnProcessAsync"),
  BdpTaskAdapter = require("../BdpTaskAdapter");

const jsonifyQstat = function(obj) {
  var results = {
    jobId: obj[0],
    user: obj[1],
    queue: obj[2],
    name: obj[3],
    sessionID: obj[4],
    NDS: obj[5],
    TSK: obj[6],
    req_memory: obj[7],
    req_time: obj[8],
    status: obj[9], 
    time: obj[10],
    comment: obj[11]
  };
  return results;
};

const __checkJobExitCode = function(jobID) {
  return new Promise((resolve, reject) => {
    const qstatProcess = child_process.spawn("qstat", ["-x", "-f", jobID]);
    let jobInfo = "";
    qstatProcess.stdout.on("readable", () => {
      const data = qstatProcess.stdout.read();
      jobInfo += data;
    });
    qstatProcess.on("error", err => {
      reject(err);
    });
    qstatProcess.on("exit", (code) => {
      if (code != 0) {
        return resolve(undefined);
      }
      jobInfo = jobInfo.trim().replace(/\n\t/g, "").split("\n");
      const results = {};
      for (var i = 1; i < jobInfo.length; i++) {
        if (jobInfo[i].indexOf(" = ") !== -1) {
          // Split key and value to 0 and 1
          var data = jobInfo[i].split(" = ");
          results[data[0].trim()] = data[1].trim();
        }
      }
      resolve(results["Exit_status"]);
    });
  });
};

class BdpPbsAdapter extends BdpTaskAdapter {
  constructor(opt) {
    opt.adapterName = "PBS Pro™";
    opt.adapterAuthor = "Chi Yang: chiyang1118@gmail.com";
    opt.dockerPath = opt.dockerPath || '/bin/docker';
    super(opt);
    this.userInfo = os.userInfo();
    this.options.stdoeMode = "watch";
    this.options.tmpfs = this.options.tmpfs || undefined;
    
  }
  async beforeExit() {
    const taskIDs = Object.keys(this.runningTasks);
    while (taskIDs.length > 0) {
      const deletingTaskIDs = taskIDs.splice(0, 5);
      const deletingPromises = [];
      for (let i = 0; i < deletingTaskIDs.length; i ++ ) {
        const deletingTaskID = deletingTaskIDs[i];
        const taskObj = this.taskLogs[deletingTaskID];
        const taskID = taskObj.taskID;
        const pbsJobID = taskObj.pid;
        if (!pbsJobID) { continue; }
        process.stderr.write(`[${new Date().toString()}] Stop job: ${pbsJobID} for task: ${taskID}` + "\n");
        const deletingPromise = (async () => {
          await spawnProcessAsync("qdel", [pbsJobID], "qdel-job-" + pbsJobID, {mode: "pipe", verbose: false, shell: true});
          await sleep(1000);
          const stdoutExists = await fse.pathExists(taskObj.stdout);
          if (stdoutExists) {
            const targetSTDout = path.join(taskObj.option.taskLogFolder, taskObj.taskName, taskID + "-stdout.txt");
            await fse.move(taskObj.stdout, targetSTDout, {overwrite: true});
          }
          const stderrExists = await fse.pathExists(taskObj.stderr);
          if (stderrExists) {
            const targetSTDerr = path.join(taskObj.option.taskLogFolder, taskObj.taskName, taskID + "-stderr.txt");
            await fse.move(taskObj.stderr, targetSTDerr, {overwrite: true});
          }
        })().catch(e => console.log(e));
        deletingPromises.push(deletingPromise);
      }
      await Promise.all(deletingPromises);
    }
  }
  async taskOverrides(taskObj) {
    const taskOption = taskObj.option;
    taskObj.nodes = isNaN(parseInt(taskOption.nodes)) ? 1 : parseInt(taskOption.nodes);
    taskObj.cpus = isNaN(parseFloat(taskOption.cpus)) ? 1 : parseFloat(taskOption.cpus);
    const { value, unit } = memHumanize(taskOption.mem, false, 0);
    taskObj.mem = value + unit.toLowerCase() + "b";
    const argTemplate = await fse.readFile(path.join(__dirname, "pbs-arg-recipe.yaml"), "utf8");
    taskObj = await super.taskOverrides(taskObj, argTemplate);
    taskObj.command = "qsub" + " " + taskObj.args.join(" ");
    return taskObj;
  }
  async processSpawn(taskID) {
    const runningTaskObj = {
      jobID: "",
      stdoutStream: null,
      stderrStream: null,
      taskEmitter: new EventEmitter()
    };
    const taskObj = this.taskLogs[taskID];
    const cwd = taskObj.option.cwd || process.cwd();
    const processObj = await spawnProcessAsync("qsub", taskObj.args, "submit-job", {mode: "memory", cwd: cwd});
    const jobID = processObj.stdout.replace(/\n/g, "");
    if (processObj.stderr) {
      process.write('StdErr when qsub: ' + processObj.stderr + '\n');
    }
    const matched = jobID.match(/^(\d+)\..*$/);
    if (!matched || !matched[1]) {
      throw 'This job has been sent, but it did NOT return the job ID.' +
            'Please try increasing the options.delayInterval value to avoid submitting too many jobs at one time.' +
            'You might also need to manually ';
    }
    taskObj.stdout = path.resolve(process.env.HOME, taskID + ".o" + matched[1]);
    taskObj.stderr = path.resolve(process.env.HOME, taskID + ".e" + matched[1]);
    taskObj.pid = jobID;
    runningTaskObj.jobID = jobID;
    return runningTaskObj;
  }

  async detectJobStatus() {
    const pbsUser = this.userInfo.username;
    try {
      const { exitCode, stdout } = await spawnProcessAsync("qstat", ["-xaw", "-u", pbsUser], "check-job-status", { mode: "memory" });
      if (exitCode != 0 || !stdout) {
        return;
      }
      const lines = stdout.split("\n");
      const jobs = {};
      for (let i = 5; i < lines.length; i++) {
        if (!lines[i]) {continue;}
        lines[i] = lines[i].replace(/^.*?\s/, function(jobname){ return jobname + "  ";});
        lines[i] = lines[i].replace(/\s[A-Z]\s/, function(status){ return "  " + status + "  ";});
        lines[i] = lines[i].trim().split(/[\s]{2,}/);
        jobs[lines[i][0]] = jsonifyQstat(lines[i]);
      }
      for (const runningTaskID of Object.keys(this.runningTasks)) {
        const taskObj = this.taskLogs[runningTaskID];
        const jobID = taskObj.pid;
        if (jobID && jobs[jobID]) {
          // Running
          const status = jobs[jobID].status;
          if (status === "F") {
            let exitCode = undefined, checkTimes = 0;
            while (exitCode == undefined && checkTimes < 5) {
              try {
                exitCode = await __checkJobExitCode(jobID);
              } catch (e) {
                console.log(e);
              }
              checkTimes++;
              if (exitCode == null) {
                await sleep(800);
              }
            }
            this.taskLogs[runningTaskID].stdout = path.join(
              taskObj.option.taskLogFolder,
              taskObj.taskName,
              runningTaskID + "-stdout.txt"
            );
            this.taskLogs[runningTaskID].stderr = path.join(
              taskObj.option.taskLogFolder,
              taskObj.taskName,
              runningTaskID + "-stderr.txt"
            );
            if (exitCode == undefined) {
              exitCode = 3;
            }
            if (exitCode == 0) {
              if (this.runningTasks[runningTaskID] && this.runningTasks[runningTaskID].taskEmitter) {
                if (this.runningTasks[runningTaskID].taskEmitter.listenerCount("finish") > 0) {
                  this.runningTasks[runningTaskID].taskEmitter.emit("finish", exitCode, null);
                }else {
                  console.log(`${runningTaskID} does not have finish event lisener`);
                }
              }
            } else {
              if (this.runningTasks[runningTaskID] && this.runningTasks[runningTaskID].taskEmitter) {
                if (this.runningTasks[runningTaskID].taskEmitter.listenerCount("error") > 0) {
                  this.runningTasks[runningTaskID].taskEmitter.emit("error", exitCode, null);
                } else {
                  console.log(`${runningTaskID} does not have error event lisener`);
                }
              }
            }
            // Checking five times (~ 5 seconds) 
          } else if (status === "E") {
            // this.runningTasks[runningTaskID].taskEmitter.emit("error");
          } else if (status === "R" || status === "Q") {
            // The task is running or in a queue
          } else {
            // Not running or in a queue, might have some errors in the job scheduler
            // Currently ignore this
            // this.runningTasks[runningTaskID].taskEmitter.emit("error");
          }
        } else {
          // may be finished
          // this.runningTasks[runningTaskID].taskEmitter.emit("error");
        }
      }
    } catch (err) {
      console.log(err);
    }
  }
}

module.exports = BdpPbsAdapter;
