const os = require("os"),
  path = require("path"),
  fse = require("fs-extra"),
  EventEmitter = require("events"),
  SSH2Promise = require("ssh2-promise"),
  sleep = require("../../utilities/sleep"),
  memHumanize = require("../../utilities/humanize-memory"),
  spawnProcessAsync = require("../../utilities/spawnProcessAsync"),
  BdpTaskAdapter = require("../BdpTaskAdapter");

const jsonifyQstat = function(obj) {
  var results = {
    jobId: obj[0],
    user: obj[1],
    queue: obj[2],
    name: obj[3],
    sessionID: obj[4],
    NDS: obj[5],
    TSK: obj[6],
    req_memory: obj[7],
    req_time: obj[8],
    status: obj[9],
    time: obj[10],
    comment: obj[11]
  };
  return results;
};

class BdpRemotePbsAdapter extends BdpTaskAdapter {
  constructor(opt) {
    opt.adapterName = "Remote PBS Pro™";
    opt.adapterAuthor = "Chi Yang: chiyang1118@gmail.com";
    super(opt);
    this.userInfo = os.userInfo();
    this.options.stdoeMode = "watch";
    this.options.dockerPath = this.options.dockerPath || "docker";
    this.currentRunningDockerName = "";
    this.options.tmpfs = this.options.tmpfs || undefined;
    /**
     * opt: {
     *  remoteUser: ''
     *  , remotePasswd: '' # If empty or ignored, i will use remoteKeyFile for authentication
     *  , remoteIP: ''
     *  , remoteKeyFile: ''
     *  , remoteMountPoint: ''
     *  , remoteHome: ''
     *  , uid:''
     *  , gid: ''
     *  , syncBeforeExit: true
     *  , syncBeforeStart: true
     *  , tmpfs: ''
     * }
     */
  }
  async rsyncData(syncPath, localPath, type, direction) {
    direction = direction || "upload";
    const remoteKeyBasename = path.basename(this.options.remoteKeyFile);
    this.currentRunningDockerName = "bdp-file-rsync";
    await spawnProcessAsync(this.options.dockerPath, ['rm', '-f', 'bdp-file-rsync'], 'stop-rsync', {
      mode: 'memory',
      verbose: true
    });
    await spawnProcessAsync(
      this.options.dockerPath,
      [
        "run",
        "--rm",
        "--name",
        this.currentRunningDockerName,
        "-v",
        '"' +
          this.options.remoteKeyFile +
          ":/key/" +
          remoteKeyBasename +
          '"',
        "-v",
        '"' + path.dirname(localPath) + ':/' + type + '"',
        "-e",
        '"USER=' + this.options.remoteUser + '"',
        "-e",
        '"IP=' + this.options.remoteIP + '"',
        "-e",
        '"REMOTE_MOUNT=' + this.options.remoteMountPoint + '"',
        "registry.gitlab.com/chiyang/bdp-rsync",
        direction,
        syncPath,
        type
      ],
      "rsync-data",
      {
        mode: "pipe",
        verbose: false
      }
    );
  }
  async checkJobExitCode(jobID) {
    let jobInfo = await this.ssh.exec("qstat -x -f " + jobID);
    jobInfo = jobInfo
      .trim()
      .replace(/\n\t/g, "")
      .split("\n");
    const results = {};
    for (var i = 1; i < jobInfo.length; i++) {
      if (jobInfo[i].indexOf(" = ") !== -1) {
        // Split key and value to 0 and 1
        var data = jobInfo[i].split(" = ");
        results[data[0].trim()] = data[1].trim();
      }
    }
    return results["Exit_status"];
  }
  async beforeStart() {
    const sshConfig = {
      host: this.options.remoteIP,
      username: this.options.remoteUser,
      identity: this.options.remoteKeyFile || undefined,
      password:
        !this.options.remoteKeyFile && this.options.remotePasswd
          ? this.options.remotePasswd
          : undefined
    };
    this.ssh = new SSH2Promise(sshConfig);
    this.sftp = new SSH2Promise.SFTP(this.ssh);
    await this.ssh.connect();
    console.log("ssh connected");
    if (this.options.syncBeforeStart) {
      if (this.options.dataFolder) {
        await this.rsyncData(
          path.basename(this.options.dataFolder) + "/",
          this.options.dataFolder,
          "data",
          "upload"
        );
      }
      if (this.options.sharedFolder) {
        console.log(this.options.sharedFolder);
        await this.rsyncData("shared/", this.options.sharedFolder, "data", "upload");
      }
      if (this.options.configFolder) {
        await this.rsyncData(path.basename(this.options.configFolder) + "/", this.options.configFolder, "config", "upload");
      }
    }
  }
  async beforeExit() {
    if (this.currentRunningDockerName && this.options.syncBeforeExit) {
      await spawnProcessAsync("docker", ['stop', this.currentRunningDockerName], "stop-runningDocker", {
        mode: "memory",
        verbose: false,
        shell: true
      })
    }
    const taskIDs = Object.keys(this.runningTasks);
    while (taskIDs.length > 0) {
      const deletingTaskIDs = taskIDs.splice(0, 5);
      const jobIDs = deletingTaskIDs.map((taskID) => {
        const jobID = this.taskLogs[taskID].pid;
        process.stderr.write("[" + new Date().toString() + "] Stop job: " + jobID + "\n");
        return jobID;
      });
      await this.ssh.exec("qdel " + jobIDs.join(" "));
    }
    await sleep(3000);
    await this.getSTDoe(taskIDs);
    if (this.options.syncBeforeExit) {
      // Sync data
      if (this.options.dataFolder) {
        console.log('Sync data');
        await this.rsyncData(path.basename(this.options.dataFolder) + "/", this.options.dataFolder, "data","download");
      }
      if (this.options.sharedFolder) {
        await this.rsyncData("shared/", this.options.sharedFolder, 'data',"download");
      }
      if (this.options.configFolder) {
        await this.rsyncData(path.basename(this.options.configFolder) + "/", this.options.configFolder, "config", "download");
      }
    }else {
      console.log('======================================================================================================');
    }
    if (this.ssh && this.ssh.close) {
      await this.ssh.close();
    }
  }
  async taskOverrides(taskObj) {
    const taskOption = taskObj.option;
    taskObj.nodes = isNaN(parseInt(taskOption.nodes))
      ? 1
      : parseInt(taskOption.nodes);
    taskObj.cpus = isNaN(parseFloat(taskOption.cpus))
      ? 1
      : parseFloat(taskOption.cpus);
    const { value, unit } = memHumanize(taskOption.mem, false, 0);
    taskObj.mem = value + unit.toLowerCase() + "b";
    const argTemplate = await fse.readFile(
      path.join(__dirname, "remote-pbs-arg-recipe.yaml"),
      "utf8"
    );
    taskObj = await super.taskOverrides(taskObj, argTemplate);
    taskObj.command = "qsub" + " " + taskObj.args.join(" ");
    return taskObj;
  }

  async processExitCallback(taskID) {
    let exitCode,
      checkTimes = 0;
    const taskObj = this.taskLogs[taskID];
    while (exitCode == undefined && checkTimes < 10) {
      try {
        exitCode = await this.checkJobExitCode(taskObj.pid);
      } catch (e) {
        console.log(e);
      }
      checkTimes++;
      await sleep(1000);
    }
    exitCode = exitCode == undefined ? 1 : exitCode;
    const stdoePrefix = taskObj.option.remoteHome + "/" + taskObj.taskID;
    const matched = taskObj.pid.match(/^(\d+)\..*$/);
    if (matched && matched[1]) {
      try {
        await this.sftp.unlink(stdoePrefix + ".o" + matched[1])
      }catch(e) { console.log(e); }
      try {
        await this.sftp.unlink(stdoePrefix + ".e" + matched[1])
      }catch(e) { console.log(e); }
    }
    return { exitCode: exitCode, signal: null };
  }

  async processSpawn(taskID) {
    const runningTaskObj = {
      jobID: "",
      stdoutStream: 0,
      stderrStream: 0,
      taskEmitter: new EventEmitter()
    };
    const taskObj = this.taskLogs[taskID];
    // const cwd = taskObj.option.cwd || process.cwd();
    const processID = await this.ssh.exec("qsub " + taskObj.args.join(" "));
    const jobID = processID.replace(/\n/g, "");
    /**
     * Note: the stdout/stderr needs to update the files each time when running detectJobStatus.
     */
    taskObj.stdout = path
      .join(
        taskObj.option.taskLogFolder,
        taskObj.taskName,
        taskID + "-stdout.txt"
      )
      .replace(/\\/g, "/");
    taskObj.stderr = path
      .join(
        taskObj.option.taskLogFolder,
        taskObj.taskName,
        taskID + "-stderr.txt"
      )
      .replace(/\\/g, "/");
    taskObj.pid = jobID;
    runningTaskObj.jobID = jobID;
    await fse.ensureFile(taskObj.stdout);
    await fse.ensureFile(taskObj.stderr);
    return runningTaskObj;
  }
  async getSTDoe(runningTaskIDs) {
    runningTaskIDs = Array.isArray(runningTaskIDs) ? runningTaskIDs : Object.keys(this.taskLogs);
    const runningTaskLength = runningTaskIDs.length;
    let processedTaskIndex = 0;
    while (processedTaskIndex < runningTaskLength) {
      const getReadStreamPromises = [];
      for (let i = 0; i < 20; i++) {
        if (processedTaskIndex >= runningTaskLength) {
          break;
        }
        const runningTaskID = runningTaskIDs[processedTaskIndex];
        if (!runningTaskID) {
          break;
        }
        const taskObj = this.taskLogs[runningTaskID];
        const runningTaskObj = this.runningTasks[runningTaskID];
        const matched = this.taskLogs[runningTaskID].pid.match(/^(\d+)\..*$/);
        const expectedSTDOEprefix = path.join(
          taskObj.option.remoteHome + "/" + taskObj.taskID
        );
        const expectedSTDOE = {
          stdout: path
            .join(expectedSTDOEprefix + ".o" + matched[1])
            .replace(/\\/g, "/"),
          stderr: path
            .join(expectedSTDOEprefix + ".e" + matched[1])
            .replace(/\\/g, "/")
        };
        getReadStreamPromises.push.apply(
          getReadStreamPromises,
          ["stdout", "stderr"].map(async stdoe => {
            try {
              const stat = await this.sftp.getStat(expectedSTDOE[stdoe]);
              if (stat.size === runningTaskObj[stdoe + "Stream"]) {
                return false;
              }
              const stdoutStream = await this.sftp.createReadStream(
                expectedSTDOE[stdoe],
                { start: runningTaskObj[stdoe + "Stream"], end: stat.size }
              );
              const writeFS = fse.createWriteStream(taskObj[stdoe], { flags: "a" });
              return new Promise(resolve => {
                stdoutStream.pipe(writeFS);
                stdoutStream.on("end", () => {
                  runningTaskObj[stdoe + "Stream"] = stat.size;
                  resolve(true);
                });
                stdoutStream.on("error", () => resolve(false));
              });
            } catch (e) {
              return false;
            }
          })
        );
        processedTaskIndex++;
      }
      try {
        await Promise.all(getReadStreamPromises);
      }catch(e) {
        console.log(e);
      }
    }
  }
  async detectJobStatus() {
    try {
      const runningTaskIDs = Object.keys(this.runningTasks);
      await this.getSTDoe(runningTaskIDs);
      const stdout = await this.ssh.exec(
        "qstat -x -w -u " + this.options.remoteUser
      );
      const lines = stdout.split("\n");
      const jobs = {};
      for (let i = 5; i < lines.length; i++) {
        if (!lines[i]) {
          continue;
        }
        lines[i] = lines[i].replace(/^.*?\s/, function myFunction(jobname) {
          return jobname + "  ";
        });
        lines[i] = lines[i].replace(/\s[A-Z]\s/, function myFunction(status) {
          return "  " + status + "  ";
        });
        lines[i] = lines[i].trim().split(/[\s]{2,}/);
        jobs[lines[i][0]] = jsonifyQstat(lines[i]);
      }
      for (const runningTaskID of runningTaskIDs) {
        const jobID = this.taskLogs[runningTaskID].pid;
        if (jobID && jobs[jobID]) {
          const status = jobs[jobID].status;
          if (status === "F") {
            this.runningTasks[runningTaskID].taskEmitter.emit('finish');
          } else if (status === "E") {
            this.runningTasks[runningTaskID].taskEmitter.emit('error');
          } else if (status === "R" || status === "Q") {
            // The task is running or in a queue
          } else {
            // Not running or in a queue, might have some errors in the job scheduler
            // Currently ignore this
            // this.runningTasks[runningTaskID].taskEmitter.emit("error");
          }
        } else {
          // this.runningTasks[runningTaskID].taskEmitter.emit("error");
        }
      }
    } catch (err) {
      console.log(err);
    }
  }
}

module.exports = BdpRemotePbsAdapter;
